using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Reflection;
using Upsilon.BaseTypes.UserData;
using Upsilon.Evaluator;
using Upsilon.Exceptions;
using Upsilon.Text;

namespace Upsilon.BaseTypes.ScriptFunction
{
    internal class ScriptMethodInfoFunction : ScriptFunction
    {
        public ScriptMethodInfoFunction(UserDataMethod method, object o, bool directTypeManipulation,
            bool passScriptReference = false, bool passScopeReference = false)
        {
            Method                 = method;
            _object                 = o;
            _directTypeManipulation = directTypeManipulation;
            _passScriptReference = passScriptReference;
            _passScopeReference = passScopeReference;

            ReturnType              = Method.ReturnType;

            if (method.GetMethods().First().Attribute != null)
            {
                var attr = method.GetMethods().First().Attribute;
                _directTypeManipulation = attr.DirectScriptManipulation;
                _passScriptReference    = attr.PassScriptReference;
                _passScopeReference     = attr.PassScopeReference;

            }
        }

        public UserDataMethod Method { get; }
        private readonly object         _object;
        private readonly bool           _directTypeManipulation;
        private readonly bool           _passScriptReference;
        private readonly bool _passScopeReference;
        public           System.Type    ReturnType { get; }

        public IEnumerable<UserDataMethod.UserDataMethodParameter> GetParameterInfo()
        {
            return Method.GetMethods().First().Parameters;
        }

        public override ScriptType Run(Diagnostics diagnostics, ScriptType[] variables, Script script, EvaluationScope scope, TextSpan span)
        {
            var objects = new List<object>();
            if (_passScriptReference)
                objects.Add(script);
            if (_passScopeReference)
                objects.Add(scope);
            objects.AddRange(variables.Select(x => (object) x).ToList() );
            object result;
            try
            {
                // If no parent object for the function was given, this is likely a lambda function, these need to be static
                if (_object == null)
                {
                    // grab the parameters, and just invoke it
                    var array = objects.ToArray();
                    var methodInfo = Method.GetMethod(ref array);
                    result = methodInfo.Invoke(null, array);
                }
                else
                {
                    // we create an array for the types of the objects
                    var convertedTypes = new System.Type[objects.Count];
                    for (var index = 0; index < objects.Count; index++)
                    {
                        // we grab the type of each parameter, and grab the underlying type if its a userdata object, as we need these
                        var arg = objects[index];
                        if (arg is IUserData ud)
                        {
                            convertedTypes[index] = ud.GetCSharpType();
                        }
                        else
                        {
                            convertedTypes[index] = arg.GetType();
                        }
                    }

                    // grab all public methods
                    var methods = _object.GetType().GetMethods(BindingFlags.Public | BindingFlags.Instance);
                    MethodInfo method = null;
                    var numberOfParametersFilled = -1;
                    // iterate over the methods on the object
                    foreach (var match in methods)
                    {
                        // if the method does not have the name we're looking for, continue
                        if (!string.Equals(match.Name, Method.Name)) continue;
                        // Use the type binder to check if the match is allowed
                        if (UpsilonBinder.Default.IsValidMatch(match, convertedTypes))
                        {
                            // grab the number of parameters on the match
                            var parCount = match.GetParameters().Length;
                            // We always try to grab the function with the most parameters
                            if (parCount > numberOfParametersFilled)
                            {
                                method = match;
                                numberOfParametersFilled = parCount;
                            }
                        }
                    }
                    // if we haven't found a method, we throw an exception
                    if (method == null)
                    {
                        throw new ScriptRuntimeException(
                            $"Can't find method {Method.Name} with parameter types {string.Join(", ", convertedTypes.Select(x => x.Name))} on type {_object.GetType().Name}",
                            span, diagnostics.ScriptString.GetSpan(span));
                    }
                    // get the method parameters
                    var parameters = method.GetParameters();
                    var arguments = objects.ToArray();
                    // if the number of parameters does not match, we need to resize our given parameters
                    if (parameters.Length != arguments.Length)
                    {
                        Array.Resize(ref arguments, parameters.Length);
                    }
                    // invoke the method
                    result = method.Invoke(_object, BindingFlags.InvokeMethod, UpsilonBinder.Default, arguments, CultureInfo.InvariantCulture);
                }
            }
            // Catch any exception inside the invocation
            catch (TargetInvocationException e)
            {
                if (e.InnerException != null) throw e.InnerException;
                throw;
            }
            if (_directTypeManipulation)
                return (ScriptType)result;
            return result.ToScriptType();

        }
    }
}