using System;
using System.Collections.Generic;
using System.Collections.Immutable;
using System.Linq;
using Upsilon.BaseTypes.ScriptTypeInterfaces;
using Upsilon.Binder;
using Upsilon.Binder.VariableSymbols;
using Upsilon.Evaluator;
using Upsilon.Text;

namespace Upsilon.BaseTypes.ScriptFunction
{
    internal class ScriptRuntimeFunction : ScriptFunction
    {
        public List<ScriptRuntimeFunctionOption> Options { get; }

        public ScriptRuntimeFunction(List<ScriptRuntimeFunctionOption> options)
        {
            Options = options;
        }

        public override ScriptType Run(Diagnostics diagnostics, ScriptType[] variables, Script script,
            EvaluationScope scope, TextSpan span)
        {
            var option = GetValidOption(variables);
            if (option == null)
            {
                throw new EvaluationException(
                    $"No valid function found");
            }
            return option.Run(diagnostics, variables, script, scope, span);
        }

        public ScriptRuntimeFunctionOption GetValidOption(object[] variables)
        {
            if (variables == null)
            {
                return Options.FirstOrDefault(x => x.Parameters.Length == 0);
            }
            foreach (var option in Options)
            {
                if (option.Parameters.Length != variables.Length)
                    continue;
                var isCompatible = true;
                for (var index = 0; index < variables.Length; index++)
                {
                    var parameter       = option.Parameters[index];
                    var parameterSymbol = ((UserDataVariableSymbol) parameter.VariableSymbol);
                    var parameterType   = variables[index].GetType();
                    if (parameterSymbol.BoundTypeDefinition != null)
                    {
                        var validSymbol =
                            parameterSymbol.BoundTypeDefinition.ValidInternalTypes.Any(validType =>
                                validType.IsAssignableFrom(parameterType));
                        if (!validSymbol)
                        {
                            isCompatible = false;
                            break;
                        }
                    }
                }
                if (!isCompatible)
                    continue;
                return option;
            }
            return null;
        }

        public ScriptRuntimeFunctionOption GetValidOption(ScriptType[] variables)
        {
            foreach (var option in Options)
            {
                if (option.Parameters.Length != variables.Length)
                    continue;
                bool isCompatible = true;
                for (var index = 0; index < variables.Length; index++)
                {
                    var callingVariable = variables[index];
                    var optionVariable = option.Parameters[index];
                    if (callingVariable.Type == BaseTypes.Type.Unknown || callingVariable.Type == BaseTypes.Type.Nil ||
                        optionVariable.Type == BaseTypes.Type.Unknown)
                        continue;
                    if (callingVariable.Type != optionVariable.Type)
                    {
                        isCompatible = false;
                        break;
                    }
                }
                if (!isCompatible)
                    continue;
                return option;
            }
            return null;
        }

        public void AddOption(ScriptRuntimeFunctionOption newOption)
        {
            var variables = newOption.Parameters;
            for (var i = 0; i < Options.Count; i++)
            {
                var option = Options[i];
                if (option.Parameters.Length != variables.Length)
                    continue;
                bool isCompatible = true;
                for (var index = 0; index < variables.Length; index++)
                {
                    var callingVariable = variables[index];
                    var optionVariable  = option.Parameters[index];
                    if (callingVariable.Type == BaseTypes.Type.Unknown || callingVariable.Type == BaseTypes.Type.Nil ||
                        optionVariable.Type  == BaseTypes.Type.Unknown)
                        continue;
                    if (callingVariable.Type != optionVariable.Type)
                    {
                        isCompatible = false;
                        break;
                    }
                }
                if (!isCompatible)
                    continue;
                Options[i] = newOption;
                break;
            }
            Options.Add(newOption);
        }


        public class ScriptRuntimeFunctionOption : IScopeOwner
        {
            public ScriptRuntimeFunctionOption(ImmutableArray<BoundVariableSymbol> parameters,
            BoundBlockStatement block, EvaluationScope scope)
            {
                Parameters = parameters;
                Block = block;
                EvaluationScope = scope;
            }

            public EvaluationScope EvaluationScope { get; }

            public BoundBlockStatement                 Block           { get; }
            public ImmutableArray<BoundVariableSymbol> Parameters      { get; }

            public ScriptType Run(Diagnostics diagnostics, ScriptType[] variables, Script script, EvaluationScope scope, TextSpan span)
            {
                var innerEvaluator = new Evaluator.Evaluator(diagnostics, EvaluationScope, script);
                if (Parameters != null)
                {
                    for (var i = 0; i < Parameters.Length; i++)
                    {
                        var parameterVariable = Parameters[i];
                        var parameterValue    = variables[i];
                        innerEvaluator.Scope.CreateLocal(parameterVariable.VariableSymbol, parameterValue);
                    }
                }
                return innerEvaluator.EvaluateNode(Block);
            }
        }
    }
}