using System;
using System.Collections.Generic;
using System.Collections.Immutable;
using System.Linq;
using System.Reflection;
using Upsilon.BaseTypes;
using Upsilon.Binder;
using Upsilon.Binder.VariableSymbols;
using Upsilon.StandardLibraries;
using Type = Upsilon.BaseTypes.Type;

namespace Upsilon.BoundTypes
{
    public class UserDataBoundTypeDefinition : BoundTypeDefinition
    {
        public string Name { get; protected set; }
        public Dictionary<string, UserDataBoundProperty> Properties { get; protected set; }

        internal UserDataBoundTypeDefinition(System.Type backingType, string name)
            : base(new TypeContainer(name), backingType)
        {
            Name = backingType.Name;
        }

        public UserDataBoundTypeDefinition(string name, Dictionary<string, UserDataBoundProperty> backingType)
            : base(new TypeContainer(name), typeof(void))
        {
            Name = name;
            Properties = backingType;
        }

        public static UserDataBoundTypeDefinition Create(System.Type backingType, string name)
        {
            var obj = new UserDataBoundTypeDefinition(backingType, name)
            {
                Properties = new Dictionary<string, UserDataBoundProperty>(),
                Name = name
            };
            var fields = backingType.GetFields().Select(x => new UserDataBoundProperty()
            {
                Name       = x.Name,
                ActualType =  x.FieldType.Name,
                Type       = x.FieldType.GetScriptType(),
            });
            foreach (var f in fields)
            {
                obj.Properties.Add(f.Name.ToLowerInvariant(), f);
            }
            var properties = backingType.GetProperties().Select(x => new UserDataBoundProperty()
            {
                Name       = x.Name,
                ActualType =  x.PropertyType.Name,
                Type       = x.PropertyType.GetScriptType(),
            });
            foreach (var f in properties)
            {
                obj.Properties.Add(f.Name.ToLowerInvariant(), f);
            }
            var methods = new Dictionary<string, UserDataBoundMethod>();
            var backingMethods = backingType.GetMethods();
            foreach (var backingMethod in backingMethods)
            {
                if (backingMethod.IsSpecialName)
                    continue;
                var methodName = backingMethod.Name;
                var attribute = backingMethod.GetCustomAttribute(typeof(ScriptFunctionAttribute));
                if (attribute is ScriptFunctionAttribute sfa )
                {
                    methodName = sfa.Name;
                }

                var parameters = backingMethod.GetParameters().Select(parameter => new UserDataBoundFunctionParameter()
                {
                    Name       = parameter.Name,
                    ActualType = parameter.ParameterType.Name,
                    Type       = StaticScope.DeriveValidTypes(parameter.ParameterType),
                    IsOptional = parameter.IsOptional
                }).ToArray();
                var option = new UserDataBoundMethodOption(backingMethod.ReturnType.GetScriptType(), parameters);
                if (methods.TryGetValue(methodName, out var func))
                {
                    func.Options.Add(option);
                }
                else
                {
                    methods.Add(methodName,
                        new UserDataBoundMethod(methodName, new List<UserDataBoundMethodOption>() {option})
                        {
                            Type = Type.Function
                        });
                }

            }
            foreach (var f in methods)
            {
                var cleanedName = f.Value.Name.ToLowerInvariant();
                // TODO: handle this better, considering overloads
                if (obj.Properties.ContainsKey(cleanedName))
                    continue;
                obj.Properties.Add(cleanedName, f.Value);
            }

            return obj;
        }
    }

    public class UserDataBoundEnumDefinition : UserDataBoundTypeDefinition
    {
        public UserDataBoundEnumDefinition(System.Type enumType, string name) : base(enumType, name)
        {
            if (!enumType.IsEnum)
                throw new Exception("Trying to bind an enum with a type that's not an enum");

            Properties = new Dictionary<string, UserDataBoundProperty>();
            var enumValues         = Enum.GetValues(enumType);

            Name = name;

            for (var i=0; i < enumValues.Length; i++)
            {
                var value = enumValues.GetValue(i);

                var valueName = value.ToString().ToLowerInvariant();
                Properties.Add(valueName, new UserDataBoundProperty()
                {
                    Name = valueName,
                    ActualType = enumType.ToString(),
                    Type = new TypeContainer(name)
                });
            }
        }

        public UserDataBoundEnumDefinition(IEnumerable<string> values, string name) : base(name, new Dictionary<string, UserDataBoundProperty>())
        {
            Properties = new Dictionary<string, UserDataBoundProperty>();
            Name = name;

            foreach (var value in values)
            {
                var valueName = value.ToLowerInvariant();
                Properties.Add(valueName, new UserDataBoundProperty()
                {
                    Name = valueName,
                    Type = new TypeContainer(name),
                    ActualType = name
                });
            }
        }
    }

    public class UserDataBoundProperty
    {
        public string Name { get; set; }
        public TypeContainer Type { get; set; }
        public virtual string ActualType { get; set; }
        public string Comment { get; set; }
    }

    public class UserDataBoundFunctionParameter : UserDataBoundProperty
    {
        public bool IsOptional { get; set; }
    }

    public class UserDataBoundMethod: UserDataBoundProperty
    {
        public override string ActualType => "Function";
        public List<UserDataBoundMethodOption> Options;

        public UserDataBoundMethod(string name, List<UserDataBoundMethodOption> options)
        {
            Name = name;
            Options = options;
        }

        public UserDataBoundMethodOption Validate(ImmutableArray<BoundExpression> callingParameters)
        {
            return Options.FirstOrDefault(x => x.ValidateParameters(callingParameters));
        }
    }

    public class UserDataBoundMethodOption
    {
        public UserDataBoundMethodOption(TypeContainer resultType, UserDataBoundFunctionParameter[] parameters)
        {
            ResultType = resultType;
            Parameters = parameters;
        }

        public TypeContainer                    ResultType { get; set; }
        public UserDataBoundFunctionParameter[] Parameters { get; set; }

        public bool ValidateParameters(ImmutableArray<BoundExpression> callingParameters)
        {
            if (callingParameters.Length   < Parameters.Count(x => !x.IsOptional)
                || callingParameters.Length > Parameters.Length)
            {
                return false;
            }

            for (var i = 0; i < callingParameters.Length; i++)
            {
                var functionParameter = Parameters[i];
                var callingParameter  = callingParameters[i];
                if (callingParameter.Type == Type.Unknown || callingParameter.Type == Type.Nil)
                    continue;

                if (!functionParameter.Type.Type.HasFlag(callingParameter.Type))
                {
                    return false;
                }

                if (functionParameter.Type.Type.HasFlag(Type.UserData))
                {
                    var variable = Binder.Binder.ResolveVariable(callingParameter, null);
                    if (variable != null && variable.TypeContainer == Type.UserData)
                    {
                        var parent =
                            (UserDataBoundTypeDefinition) ((UserDataVariableSymbol) variable).BoundTypeDefinition;
                        if (functionParameter.ActualType != null &&
                            !string.Equals(functionParameter.ActualType, parent.Name, StringComparison.InvariantCultureIgnoreCase))
                        {
                            return false;
                        }
                    }
                }
            }

            return true;
        }

    }

}